# Electre 1S implementation
# parameters:
#   PM             - performance matrix
#   w              - weights of the criteria
#   Q              - indifference threshold
#   P              - preference thresholds
#   V              - veto thresholds
#   minmaxcriteria - vector with identification of direction of the criteria
#   lambda         - cut-off criterion
Electre_1S <- function(PM, w, Q, P, V, minmaxcriteria = 'max', lambda = 0.5){
  ## input data consistency check
  if (is.null(dim(PM))) stop("less than 2 criteria or 2 alternatives")
  if (!(is.matrix(PM) || (is.data.frame(PM)))) stop("wrong performanceMatrix, should be a matrix or a data frame")
  if (!is.numeric(unlist(PM))) stop('Only numeric values in performance matrix expected')
  PM   <- util_pm_minmax(PM, minmaxcriteria) #validate minmax and invert scales if necessary
  ncri <- ncol(PM)  #no. of criteria
  nalt <- nrow(PM)  #no. of alternatives
  alt  <- rownames(PM)
  cri  <- colnames(PM)
  if (!is.vector(w)) stop("criteriaWeights should be a vector")
  if (!is.numeric(w)) stop("criteriaWeights should be a numeric vector")
  if (ncri != length(w)) stop("length of criteriaWeights should be checked")
  if (!is.vector(P)) stop("preference thresholds should be a vector")
  if (!is.numeric(P)) stop("preference thresholds should be a numeric vector")
  if (length(P) != ncri) stop("number of preference thresholds != number of criteria")
  if (!(is.vector(Q))) stop("indifference thresholds should be a vector")
  if (!is.numeric(Q)) stop("indifference thresholds be a numeric vector")
  if (length(Q) != ncri) stop("number of indifference thresholds != number of criteria")
  if (!is.vector(V)) stop("veto thresholds should be a vector")
  if (!is.numeric(V)) stop("veto thresholds be a numeric vector")
  if (length(V) != ncri) stop("number of veto thresholds != number of criteria")
  if (!is.null(lambda) && !is.numeric(lambda)) stop("if lambda is set, it must be numeric value")
  if (lambda < 0.5 || lambda > 1) stop("Lambda value must be in interval <0.5;1>")
  prob <- 0
  for(j in 1:ncri){ # check consistency of the thresholds
    # 0 <= Qj <= Pj < Vj
    if(0 > Q[j]) {
      prob <- prob + 1
      print(cat("Q > 0 for criterion: ", cri[j]))
    }
    if(Q[j] > P[j]){
      prob <- prob + 1
      print(cat("Q > P for criterion: ", cri[j]))
    }
    if(P[j] >= V[j]){
      prob <- prob + 1
      print(cat("P >= V for criterion: ", cri[j]))
    }
    if(prob > 0) stop("Problem with consistency of threshold, see notes above.")
  }
  ## End of checking the validity of the "inputs"

  #global_concordance = global_concordance_matrix(dataset, Q = Q, P = P, W = W)
  # cj <- list() # list of criteria's concordance indexes
  # for(j in 1:ncri){ # for all criteria
  #   cj_ab <- matrix(0, nalt, nalt) #concordance index of the criterion j
  #   rownames(cj_ab) <- alt
  #   colnames(cj_ab) <- alt
  #   diag(cj_ab) <- 1
  #   for(i in 1:nalt){
  #     for(k in 1:nalt){
  #       if(i != k){
  #         #compute concordance index of the criterion j
  #         if(P[j] < PM[k,j] - PM[i,j]){
  #           cj_ab[i,k] <- 0
  #         }else if(PM[k,j] - PM[i,j] <= Q[j]){
  #           cj_ab[i,k] <- 1
  #         }else{
  #           cj_ab[i,k] <- (P[j] - PM[k,j] + PM[i,j])/(P[j] - Q[j])
  #         }
  #       }
  #     }
  #   }
  #   cj[j] <- cj_ab
  # }

  #overall concordance index oci
  t <- Electre3_ConcordanceIndex(PM, P, Q, w)
  oci <- t$ConcordanceMatrix
  # oci <- matrix(0, nalt, nalt)
  # rownames(oci) <- alt
  # colnames(oci) <- alt
  # diag(oci) <- 1
  # for(j in 1:ncri) oci <- oci + w[j] * cj[[j]]
  # oci <- oci/sum(w)

  # discordance index of the criteria
  # dj(ai,ak) = 0, if gj(ak) - gj(ai) < Vj (ai,ak) - qj (ai,ak) * ((1-c(a,b))/(1-c))
  # else dj(ai,ak) = 1
  # dj <- list() # discordance inndex of the criteria
  w_sum <- 1
  di <- matrix(0, nalt, nalt)
  rownames(di) <- alt
  colnames(di) <- alt
  diag(di) <- 1
  for(i in 1:nalt){
    dj_ab <- matrix(0, nalt, nalt)
    rownames(dj_ab) <- alt
    colnames(dj_ab) <- alt
    for(j in 1:ncri){
      for(k in 1:ncri){
        if(j != k){
          if (w[k]/sum(w) != 0) w_sum = (w[k]/sum(w))
          if (PM[j,k] - PM[i,k] >= (V[k] - Q[k]*((1 - oci[i,j] - (w_sum)) / (1 - lambda - (w_sum))))) di[i,j] = 1
        }
      }
    }
  }

  # credibility matrix
  cred <- matrix(0, nalt, nalt)
  rownames(cred) <- alt
  colnames(cred) <- alt
  for(i in 1:nalt){
    for(j in 1:nalt){
      if (i != j && oci[i,j] >= lambda && di[i,j] == 0)  cred[i,j] = 1
    }
  }

  G  <- simplify(graph_from_adjacency_matrix(cred)) #remove loops
  AM <- Graph2AdjancancyMatrix(G, alt)

  # establish kernel of the decision
  t <- ELECTRE1_Kernel(AM)
  kernel <- t$kernel
  dominated <- t$dominated
  graph <- t$praph

  # output to screen
  print('Concordance matrix')
  print(oci)
  print('discordance matrix')
  print(di)
  print('Credibility matrix')
  print(AM)
  print('kernel')
  print(kernel)
  print('Dominated alternatives')
  print(dominated)
  print('Warning - experimental implementation, do not use in production environment')

  out <- list(
    ConcordanceMatrix = oci,
    DiscordanceIndex = di,
    CredibilityIndex = AM,
    Kernel = kernel,
    Dominated = dominated,
    graph = graph
  )
  return(out)
}
