# parameters
#   PM - performance matrix criteria in columns and alternatives in rows
#   w  - weights
#   c-, c0 and c+ fuzzy concordance threshold
#   c_minus = 0.65
#   c_zero = 0.75
#   c_plus = 0.85
#   d-, d+ discordance threshold
#   d_minus = 0.25
#   d_plus = 0.5
# solution for the problem is dominance matrix which can be visualized into network graph
Electre_2 <- function(PM, w,
                      minmaxcriteria = 'max',
                      c_minus = 0.65, c_zero = 0.75, c_plus = 0.85,
                      d_minus = 0.25, d_plus = 0.5){

  #require("igraph")

  ## check validity of the objects manipulated by the current function
  # if there are less than 2 criteria or 2 alternatives, there is no MCDA problem
  if (is.null(dim(PM))) stop("less than 2 criteria or 2 alternatives")
  if (!(is.matrix(PM) || (is.data.frame(PM)))) stop("wrong performanceMatrix, should be a matrix or a data frame")
  if (!is.numeric(unlist(PM))) stop('Only numeric values in performance matrix expected')
  if (!(is.vector(w))) stop("criteriaWeights should be a vector")
  if (!is.numeric(w)) stop("criteriaWeights should be a numeric vector")
  if (ncol(PM)!=length(w)) stop("length of criteriaWeights should be checked")
  PM <- util_pm_minmax(PM, minmaxcriteria) #validate minmax and invert scales if neccessary
  if(!is.numeric(c_minus) || c_minus < 0 || c_minus > 1) stop("Concordance threshold (minimum) out of bounds")
  if(!is.numeric(c_zero) || c_zero < 0 || c_zero > 1) stop("Concordance threshold (middle) out of bounds")
  if(!is.numeric(c_plus) || c_plus < 0 || c_plus > 1) stop("Concordance threshold (maximum) out of bounds")
  if(!is.numeric(d_minus) || d_minus < 0 || d_minus > 1) stop("Discordance threshold (minimum) out of bounds")
  if(!is.numeric(d_plus) || d_plus < 0 || d_plus > 1) stop("Discordance threshold (maximum) out of bounds")
  if(c_minus > c_zero) stop("Concordance level c- > c0 (must be c- <= c0)")
  if(c_zero > c_plus) stop("Concordance level c0 > c+ (must be c0 <= c+)")
  if(d_minus > d_plus) stop("Discordance level d- > d+ (must be d- <= d+)")
  ## End of checking the validity of the "inputs"

  nalt  <- nrow(PM)  #no. of alternatives
  ncri  <- ncol(PM)  #no. of criteria
  alt   <- rownames(PM)
  alt_D <- rep(0, times=nalt)
  alt_A <- rep(0, times=nalt)

  #computes preference ratio P+/P- and constructs strong or weak outranking matrixes
  # in dominance_s or dominance_w
  preference_ratio <- function(dom, PM){
    nalt <- length(rownames(PM))
    ncri <- length(colnames(PM))
    S <- matrix(data = 0, nrow = nalt, ncol = nalt)
    for(i in 1:nalt){
      for(j in 1:nalt){
        if(dom[i,j] == 1){
          #P+(a,b)/P-(a,b) >= 1
          w_plus <- 0
          w_minus <- 0
          for(k in 1:ncri){
            if(PM[i,k] > PM[j,k]){  #P+(a,b)
              w_plus <- w_plus + w[k]
            }else if(PM[i,k] < PM[j,k]){  #P-(a,b)
              w_minus <- w_minus + w[k]
            }
          }
          if(w_minus == 0 || w_plus/w_minus >= 1) S[i,j] <- 1
        }
      }
    }
    return(S)
  }

  #function for ranking
  # dom_s - simplified graph of strong preferences
  # dom_d - simplified graph of weak preferences
  # alt   - alternatives
  ranking <- function(dom_s, dom_w, alt){
    #py code: dominance = np.clip(2*dominance_s + dominance_w, 0, 2) #since dom_s and _w should consist of 0/1:
    dominance <- 2 * dom_s + dom_w
    dominance[dominance == 3] <- 2
    colnames(dominance) <- alt
    rownames(dominance) <- alt
    y <- alt
    nalt <- length(alt)
    rank <- list()
    while (length(y) > 0){
      d <- NULL
      u <- NULL
      b <- NULL
      a <- NULL
      for(j in 1:nalt){
        check_d <- 0
        for(i in 1:nalt){
          if (dominance[i, j] == 2) check_d <- check_d + 1
        }
        #all options not strongly outranked by any other option form the set D.
        if (check_d == 0 && sum(dominance[,j]) != -nalt) d[length(d)+1] <- alt[j]
      }
      idx <- match(d, alt)  #get index of alternatives in D set
      for(m in 1:length(d)){
        for(n in 1:length(d)){
          if (dominance[idx[m], idx[n]] == 1){
            print(cat("Weak relation between ", idx[m], " and ", idx[n])) #ladění
            #Those options within D that are linked to each other by a weak outranking
            #relationship constitute the set U
            if(!(alt[idx[m]] %in% u)) u[length(u)+1] <- alt[idx[m]]
            if(!(alt[idx[n]] %in% u)) u[length(u)+1] <- alt[idx[n]]
          }
        }
      }
      idx <- match(u, alt)
      if(length(idx) > 0){
        for(m in 1:length(u)){
          check_b <- 0
          for(n in 1:length(u)){
            if (dominance[idx[n], idx[m]] == 1) check_b = check_b + 1
          }
          if (check_b == 0){
            #The set B consists of all options from U not weakly outranked by any other from within U.
            if(!(alt[idx[m]] %in% b)) b[length(b)+1] <- alt[idx[m]]
          }
        }
      }
      # The single option or group of equally ranked options is defined by the union of the sets (D - U) and B
      a <- d[!(d %in% u)]
      if(length(b) > 0) a <- append(a, b)
      if(length(a) > 0) rank <- append(rank, a)
      idx <- match(a, alt)
      y <- y[!(y %in% a)]
      if(length(idx) > 0) {
        for(j in 1:length(idx)){
          dominance[, idx[j]] <- -1
          dominance[idx[j],]  <- -1
        }
      }
    }
    return(rank)
  }

  cm <- ELECTRE_ConcordanceMatrix(PM, w)  #Concordance matrix
  dm <- ELECTRE_DiscordanceMatrix(PM) #discordance matrix

  #dominance matrix (preference matrix) - basic
  dominance_s <- matrix(data = 0, nrow = nalt, ncol = nalt)
  dominance_w <- matrix(data = 0, nrow = nalt, ncol = nalt)
  for(i in 1:nalt){
    for(j in 1:nalt){
      if(cm[i,j] >= cm[j,i] && i != j) {
        #strong outranking relation
        if(((cm[i,j] >= c_plus) && (dm[i,j] <- d_plus)) || ((cm[i,j] >= c_zero) && (dm[i,j] <= d_minus))){
          dominance_s[i,j] = 1
        }
        #weak outranking relation
        if((cm[i,j] >= c_zero && dm[i,j] <= d_plus) || (cm[i,j] >= c_minus && dm[i,j] <= d_minus)){
          dominance_w[i,j] = 1
        }
      }
    }
  }

  #P+/P- ratio evaluation
  S_strong  <- preference_ratio(dominance_s, PM)
  S_weak    <- preference_ratio(dominance_w, PM)
  G_strong  <- simplify(graph_from_adjacency_matrix(S_strong)) #remove loops
  AM_strong <- Graph2AdjancancyMatrix(G_strong, alt)
  G_weak    <- simplify(graph_from_adjacency_matrix(S_weak))
  AM_weak   <- Graph2AdjancancyMatrix(G_weak, alt)
  #ranking
  rank_A   <- ranking(AM_strong, AM_weak, alt)  #1st total preorder V1
  print("first preorder")
  print(rank_A)
  #The second pre-order is obtained using the inverse ranking procedure. The same basic
  #algorithm as above is utilised, with the following modifications.
  #invert the direction of the strong and weak outranking graphs GF, and Gf,
  #the ranks obtained from the inverted graphs using the direct procedure is adjusted as follows:
  #'inverse procedure' rank (r2(a»=1+(the number of ranks) (r'2(a)max) -'direct procedure' ranking (r'2(a))
  rank_D   <- rev(ranking(t(AM_strong), t(AM_weak), alt)) #2nd total preorder V2
  #final partial pre-order
  pref <- matrix(data = 0, nrow = nalt, ncol = nalt)
  incompar <- matrix(data = 0, nrow = nalt, ncol = nalt)
  for(i in 1:nalt){
    r1_a <- match(alt[i], rank_A)
    r1_d <- match(alt[i], rank_D)
    for(j in 1:nalt){
      if(i != j){
        r2_a <- match(alt[j], rank_A)
        r2_d <- match(alt[j], rank_D)
        #if a is preferred to b in 2 complete pre-orders, then this will also be the case in final pre-order
        if(r1_a > r2_a && r1_d > r2_d) pref[i,j] <- 1
        #ifa has an equivalent ranking to b in the one of the complete pre-orders, but is
        #preferred to b in the other, then a precedes b in the final pre-order
        if((r1_a == r2_a && r1_d > r2_d) || (r1_a > r2_a && r1_d == r2_d)) pref[i,j] <- 1
        #if a is preferred to b in the first complete pre-order, but b is preferred to a in the
        #second, then the two options are incomparable in the final pre-order.
        if((r1_a > r2_a && r1_d < r2_d) || (r1_a < r2_a && r1_d > r2_d)) incompar[i,j] <- 1
      }
    }
  }
  colnames(pref) <- alt
  rownames(pref) <- alt
  colnames(incompar) <- alt
  rownames(incompar) <- alt
  graph <- plot.prefM(pref)
  rank_F <- rowSums(pref)
  names(rank_F) <- alt
  rank_F <- sort(rank_F) + 1

  print('Concordance Matrix')
  print(cm)
  print('Discordance Matrix')
  print(dm)
  print('Strong outranking relation')
  print(AM_strong)
  print('weak outranking relation')
  print(AM_weak)
  print('First total pre-order, V1')
  print(rank_A)
  print('Secont total pre-order, V2')
  print(rank_D)
  print('Preference matrix (final partial preorder)')
  print(pref)
  print('Incomparable options')
  print(incompar)
  print('Final partial preorder')
  print(rank_F)

  out <- list(
    PerformanceMatrix   = PM,
    ConcordanceMatrix   = t(cm),
    DiscordanceMatrix   = t(dm),
    StrongOutranking    = AM_strong,
    WeakOutranking      = AM_weak,
    firstTotalPreorder  = rank_A,
    secondTotalPreorder = rank_D,
    finalPreorderMatrix = pref,
    incomparableAlternatives = incompar,
    graphResult         = graph,
    finalPreorder       = rank_F)
  return(out)
}
