# parameters
#   PM - performance matrix criteria in columns and alternatives in rows
#   w  - weights
#   P  - preference threshold
#   Q  - indifference threshold
#   V  - veto threshold
#   alpha - first discrimination threshold
#   beta  - second discrimination threshold
#   VERBOSE - in verbose mode the function prints out higher amount of information on computation
# solution for the problem is dominance matrix which can be visualized into network graph
Electre_3 <- function(PM, w, P, Q, V, minmaxcriteria = 'max',
                      alpha = 0.3, beta = 0.15, VERBOSE = F){
  ## check validity of the objects manipulated by the current function
  # if there are less than 2 criteria or 2 alternatives, there is no MCDA problem
  if (is.null(dim(PM))) stop("less than 2 criteria or 2 alternatives")
  if (!(is.matrix(PM) || (is.data.frame(PM)))) stop("wrong performanceMatrix, should be a matrix or a data frame")
  if (!is.numeric(unlist(PM))) stop('Only numeric values in performance matrix expected')
  if (!is.vector(w)) stop("criteriaWeights should be a vector")
  if (!is.numeric(w)) stop("criteriaWeights should be a numeric vector")
  PM <- util_pm_minmax(PM, minmaxcriteria) #validate minmax and invert scales if necessary
  ncri <- ncol(PM)  #no. of criteria
  if (ncri != length(w)) stop("length of criteriaWeights should be checked")
  if (!is.vector(P)) stop("preference thresholds should be a vector")
  if (!is.numeric(P)) stop("preference thresholds should be a numeric vector")
  if (length(P) != ncri) stop("number of preference thresholds != number of criteria")
  if (!(is.vector(Q))) stop("indifference thresholds should be a vector")
  if (!is.numeric(Q)) stop("indifference thresholds be a numeric vector")
  if (length(Q) != ncri) stop("number of indifference thresholds != number of criteria")
  if (!is.vector(V)) stop("veto thresholds should be a vector")
  if (!is.numeric(V)) stop("veto thresholds be a numeric vector")
  if (length(V) != ncri) stop("number of veto thresholds != number of criteria")
  if (all(P == Q)) stop("Preference and indifference thresholds are defined same. Revisit to distinguish them or use another method.")
  if (!is.numeric(alpha)) stop("alpha discrimination threshold should be a number")
  if (!is.numeric(beta)) stop("beta discrimination threshold should be a number")
  if (alpha < beta) stop("alpha shoud be > then beta (discrimination threshold)")
  ## End of checking the validity of the "inputs"

  nalt  <- nrow(PM)  #no. of alternatives
  alt   <- rownames(PM)
  alt_D <- rep(0, times=nalt)
  alt_A <- rep(0, times=nalt)

  t <- Electre3_ConcordanceIndex(PM, P, Q, w)
  cm <- t$ConcordanceMatrix
  dj <- list()  #discordance matrix for criteria (j)
  for(k in 1:ncri){      #index for criteria
    d  <- matrix(data = 0, nrow = nalt, ncol = nalt)  #temp var for dj
    colnames(d) <- alt
    rownames(d) <- alt
    for(i in 1:nalt){    #index for a
      for(j in 1:nalt){  #index for b
        #discordance matrix
        # Dj(a,b) = 0 if PMj(b) <= PMj(a) + Pj(a)
        # Dj(a,b) = 1 if PMj(b) > PMj(a) + Vj(a)
        # otherwise Dj(a,b) = (PMj(b)-PMj(a) - Pj(a))/ (Vj(a) - Pj(a))
        if(PM[j,k] <= PM[i,k] + P[k]){
          d[i,j] <- 0
        }else if(PM[j,k] > PM[i,k] + V[k]){
          d[i,j] <- 1
        }else{
          if(V[k] - P[k] != 0){
            d[i,j] <- (PM[j,k] - PM[i,k] - P[k])/(V[k] - P[k])
          }else{
            d[i,j] <- 1
          }
        }
      }
    }
    dj[[k]]  <- d
  }

  # degree of credibility
  # S(a,b) = C(a,b) if Dj(a,b) <= C(a,b)
  # otherwise S(a,b) = C(a,b) * PRODUCT{for j in J}((1-Dj(a,b))/(1-C(a,b)))
  # where J is set of criteria for which Dj(a,b) > C(a,b)
  sm  <- matrix(data = 0, nrow = nalt, ncol = nalt)  # credibility matrix
  colnames(sm) <- alt
  rownames(sm) <- alt
  for(i in 1:nalt){
    for(j in 1:nalt){
      t <- 1 #temporary var. for computation of product part of the equation
      for(k in 1:ncri){
        dj_temp <- dj[[k]]
        if(dj_temp[i,j] > cm[i,j]) t <- t * ((1-dj_temp[i,j])/(1-cm[i,j]))
      }
      sm[i,j] <- cm[i,j] * t
    }
  }
  distilationdesc <- Electre_desc_dist(sm) #descending distilation
  rank_D <- rankDF(distilationdesc)
  distilationasc <- Electre_asc_dist(sm)       #ascending distilation
  rank_A <- rankDF(distilationasc)
  rank_P <- pre_order_matrix(rank_D, rank_A, alt)
  final_ranking <- finalRanking(alt, rank_P)
  #construct adjancancy matrix
  names_matrank_adj=final_ranking[,1]
  matrank_adj <- matrix(data = 0, nrow = nalt, ncol = nalt)
  rownames(matrank_adj)=names_matrank_adj
  colnames(matrank_adj)=names_matrank_adj
  PIR <- c('P+', 'I', 'R')
  for (i in 1:nalt){
    for (j in 1:nalt){
      if(final_ranking[j,3]-final_ranking[i,3] == 1 && rank_P[final_ranking[i,1],final_ranking[j,1]] %in% PIR){
        matrank_adj[i,j] = 1
      }
    }
  }
  graph <- plot.prefM(matrank_adj)

  print('Concordance matrix')
  print(t(cm))
  print('Discordance matrixes of criteria')
  print(dj)
  print('Credibility matrix')
  print(sm)
  print('First pre-order (down)')
  print(rank_D)
  print('First pre-order (up)')
  print(rank_A)
  print('total')
  print(rank_P)
  print('Adjancancy matrix')
  print(matrank_adj)
  print('Final ranking')
  print(final_ranking)

  out <- list(
    PerformanceMatrix = PM,
    ConcordanceMatrix = cm,
    DiscordanceMatrixCriteria = dj,
    CredibilityMatrix = sm,
    rank_D = rank_D,
    rank_A = rank_A,
    rank_P = rank_P,
    adjancancyMatrix = matrank_adj,
    graph = graph,
    final_ranking = final_ranking
  )
  return(out)
}
