# parameters
#   PM - performance matrix criteria in columns and alternatives in rows
#   P  - preference threshold
#   Q  - indifference threshold
#   V  - veto threshold
#
# solution for the problem is dominance matrix which can be visualized into network graph
Electre_4 <- function(PM, P, Q, V, minmaxcriteria = 'max'){
  #require("igraph")

  ## check validity of the objects manipulated by the current function
  # if there are less than 2 criteria or 2 alternatives, there is no MCDA problem
  if (is.null(dim(PM))) stop("less than 2 criteria or 2 alternatives")
  if (!(is.matrix(PM) || (is.data.frame(PM)))) stop("wrong performanceMatrix, should be a matrix or a data frame")
  if (!is.numeric(unlist(PM))) stop('Only numeric values in performance matrix expected')
  PM <- util_pm_minmax(PM, minmaxcriteria) #validate minmax and invert scales if necessary
  ncri  <- ncol(PM)  #no. of criteria
  if (!is.vector(P)) stop("preference thresholds should be a vector")
  if (!is.numeric(P)) stop("preference thresholds should be a numeric vector")
  if (length(P) != ncri) stop("number of preference thresholds != number of criteria")
  if (!(is.vector(Q))) stop("indifference thresholds should be a vector")
  if (!is.numeric(Q)) stop("indifference thresholds be a numeric vector")
  if (length(Q) != ncri) stop("number of indifference thresholds != number of criteria")
  if (!is.vector(V)) stop("veto thresholds should be a vector")
  if (!is.numeric(V)) stop("veto thresholds be a numeric vector")
  if (length(V) != ncri) stop("number of veto thresholds != number of criteria")
  if (all(P == Q)) stop("Preference and indifference thresholds are defined same. Revisit to distinguish them or use another method.")
  ## End of checking the validity of the "inputs"

  nalt <- nrow(PM) #no. of alternatives
  alt  <- rownames(PM) #names of alternatives

  #count outranking relations
  mp <- matrix(0, nalt, nalt) #number of criteria for which option b is strictly preferred to a
  mq <- matrix(0, nalt, nalt) #number of criteria for which option b is weakly preferred to a
  mi <- matrix(0, nalt, nalt) #number of criteria for which option b is judged indifferent to a
  mo <- matrix(0, nalt, nalt) #mo(b,a) = mo(a,b) : the number of criteria on which options a and b perform identically
  veto  <- matrix(0, nalt, nalt)
  for(i in 1:nalt){
    for(j in 1:nalt){
      if (i != j){
        for(k in 1:ncri){
          if (PM[j,k] - PM[i,k] > P[k]) mp[i,j] <- mp[i,j] + 1
          if (PM[j,k] - PM[i,k] > Q[k] && PM[j,k] - PM[i,k] <= P[k]) mq[i,j] <- mq[i,j] + 1
          if (PM[j,k] - PM[i,k] >= -Q[k] && PM[j,k] - PM[i,k] <= Q[k] && PM[j,k] - PM[i,k] > 0) mi[i,j] = mi[i,j] + 1
          if (PM[j,k] - PM[i,k] == 0) mo[i,j] = mo[i,j] + 1
          if (PM[j,k] - PM[i,k] >= V[k]) veto[i,j] = veto[i,j] + 1
        }
      }
    }
  }

  #credibility matrix
  cred_matrix <- matrix(0, nalt, nalt)
  rownames(cred_matrix) <- alt
  colnames(cred_matrix) <- alt
  for(i in 1:nalt){
    for(j in 1:nalt){
      if (i != j){
        # original pyDecisions implementation
        # if (mp_ab[i,j] == 0) cred_matrix[i,j] = 0.4
        # if (mp_ab[i,j] == 0 || mp_ab[i,j] == 1 && mp_ba[i,j] >= ncri/2 && veto[i,j] >= 1) cred_matrix[i,j] = 0.2
        # if (mp_ab[i,j] == 0 && mq_ab[i,j] <= mq_ba[i,j] + mp_ba[i,j]) cred_matrix[i,j] = 0.6
        # if (mp_ab[i,j] == 0 && mq_ab[i,j] <= mq_ba[i,j] && mq_ab[i,j] + mi_ab[i,j] <= mi_ba[i,j] + mq_ba[i,j] +  mp_ba[i,j]) cred_matrix[i,j] = 0.8
        # if (mp_ab[i,j] + mq_ab[i,j] == 0 && mi_ab[i,j] < mi_ba[i,j] + mq_ba[i,j] +  mp_ba[i,j]) cred_matrix[i,j] = 1.0
        if (mp[i,j] + mq[i,j] == 0 && mi[i,j] < mi[j,i] + mq[j,i] +  mp[j,i]){
          #Quasi-dominance Sq: b outranks a with quasi-dominance if
          #b Sq a <=> mp(a,b) + mq(a,b) = 0 and mi(a,b) < mi(b,a) + mq(b,a) + mp(b,a)
          cred_matrix[i,j] = 1.0
        } else if(mp[i,j] == 0 && mq[i,j] <= mq[j,i] && mq[i,j] + mi[i,j] <= mi[j,i] + mq[j,1] +  mp[j,i] + 1) {
          # Canonical dominance Sc: b outranks a with canonical dominance if
          # b Sc a <=> mp(a,b) = 0, and mq(a,b) <= mq(b,a), and mq(a,b) + mi(a,b) <= mi(b,a) + mq(b,a) + mp(b,a) + 1
          cred_matrix[i,j] = 0.8
        } else if (mp[i,j] == 0 && mq[i,j] <= mq[j,i] + mp[j,i]) {
          # Pseudo-dominance Sp: b outranks a with pseudo-dominance if
          # b Sp a <=> mp(a,b) = 0, and II1q(a,b) S; II1q(b,a) + mp(b,a)
          cred_matrix[i,j] = 0.6
        } else if (mp[i,j] == 0) {
          # Sub-dominance Ss: b outranks a with sub-dominance if
          # b Ss a: mp(a,b) == 0
          cred_matrix[i,j] = 0.4
        } else if (mp[i,j] == 1 && mp[j,i] >= ncri/2) { # && veto[i,j] >= 1)
          # Veto dominance Sv: b outranks a with veto-dominance if
          # if mp(a,b) = 1
          # b Sv a: mp(b,a) >= m/2, and gj(b) + vj[gj(b)] >= gj(a)
          le <- T
          for(k in 1:ncri){ #test across all criteria
            if(PM[j,k] + V[k] < PM[i,k]) {
              le <- F
              break
            }
          }
          if(le) cred_matrix[i,j] = 0.2
        } else if(mp[i,j] == 0){
          # Veto dominance Sv: b outranks a with veto-dominance if
          # if mp(a,b) = 0
          cred_matrix[i,j] = 0.2
        } # else sm = 0 <- implied as sm is initialized to 0
      }
    }
  }

  print(cred_matrix)
  distilationdesc <- Electre_desc_dist(cred_matrix)     #descending distillation
  print(distilationdesc)
  rank_D <- rankDF(distilationdesc)
  print(rank_D)
  distilationasc <- Electre_asc_dist(cred_matrix)       #ascending distillation
  rank_A <- rankDF(distilationasc)
  print(rank_A)
  rank_P <- pre_order_matrix(rank_D, rank_A, alt)
  final_ranking <- finalRanking(alt, rank_P)
  #construct adjancancy matrix
  names_matrank_adj=final_ranking[,1]
  matrank_adj <- matrix(data = 0, nrow = nalt, ncol = nalt)
  rownames(matrank_adj)=names_matrank_adj
  colnames(matrank_adj)=names_matrank_adj
  PIR <- c('P+', 'I', 'R')
  for (i in 1:nalt){
    for (j in 1:nalt){
      if(final_ranking[j,3]-final_ranking[i,3] == 1 && rank_P[final_ranking[i,1],final_ranking[j,1]] %in% PIR){
        matrank_adj[i,j] = 1
      }
    }
  }
  graph <- plot.prefM(matrank_adj)

  print('Credibility matrix')
  print(cred_matrix)
  print('First pre-order (down)')
  print(rank_D)
  print('First pre-order (up)')
  print(rank_A)
  print('total')
  print(rank_P)
  print('Adjancancy matrix')
  print(matrank_adj)
  print('Final ranking')
  print(final_ranking)

  out <- list(
    PerformanceMatrix = PM,
    CredibilityMatrix = cred_matrix,
    rank_D = rank_D,
    rank_A = rank_A,
    rank_P = rank_P,
    adjancancyMatrix = matrank_adj,
    graph = graph,
    final_ranking = final_ranking
  )
  return(out)
}
