# Function provides basic computations for PROMETHEE I, II and III methods. It computes positive and negative
# preference flows which are used as basis of recomendation formulation.
#
# parameters
#   PM - performance matrix
#   preferenceFunction - vector of preference functions types to derive preference when comparing PM
#   w - weights
#   indifferenceTreshold list of indifference thresholds
#   prefferenceThreshold list of prefference thresholds
#   intermediateThreshold lish of intermediate thresholds for Gaussian function type
PROMETHEE <- function(PM, preferenceFunction, w, indifferenceTreshold = NULL, prefferenceThreshold = NULL,
                      intermediateThreshold = NULL){
  ## check validity of the objects manipulated by the current function
  # if there are less than 2 criteria or 2 alternatives, there is no MCDA problem
  if (is.null(dim(PM))) stop("less than 2 criteria or 2 alternatives")
  if (!(is.matrix(PM) || (is.data.frame(PM)))) stop('wrong performanceMatrix, should be a matrix or a data frame')
  if (!is.numeric(unlist(PM))) stop('Only numeric values in performance matrix expected')
  ncri <- ncol(PM)  #no. of criteria
  qj   <- indifferenceTreshold
  pj   <- prefferenceThreshold
  sj   <- intermediateThreshold
  fTypes <- c('default', 'U-shape', 'V-shape', 'level', 'linear', 'Gaussian')
  indifTypes <- c('U-shape', 'level', 'linear')
  indifTypesVector <- rep(F, times=ncri)
  prefTypes <- c('V-shape', 'level', 'linear')
  prefTypesVector <- rep(F, times=ncri)
  if (!(is.vector(preferenceFunction))) stop('preferenceFunction must be a vector')
  for(i in 1:ncri){ #check for consistency of preference function
    if(!(preferenceFunction[i] %in% fTypes)) stop('unsupported type of preference function used only: default, U-shape, V-shape, level, linear, Gaussian supported.')
    if(preferenceFunction[i] %in% indifTypes) indifTypesVector[i] <- T
    if(preferenceFunction[i] %in% prefTypes) prefTypesVector[i] <- T
  }
  if (T %in% indifTypesVector){ # indifference threshold required, check its consistency
    if(is.null(qj)) stop('some choosen preference functions need to have set indifference threshold.')
    if (!is.numeric(qj)) stop('indifferenceTreshold must be a numeric vector')
    if (length(qj) != ncri) stop('number of elements in indifferenceTreshold must be equal to number of criteria (set value to 0 for criteria with preference function that does not require this parameter).')
    for(i in 1:ncri){
      if (indifTypesVector[i] && qj[i] < 0){
        msg <- cat('indifference threshold for criterion ', i, ' set < 0, only non negative values accepted.')
        stop(msg)
      }
    }
  }
  if (T %in% prefTypesVector){ # prefference threshold required, check its consistency
    if(is.null(pj)) stop('some choosen preference functions need to have set prefference threshold.')
    if (!is.numeric(pj)) stop('prefferenceThreshold must be a numeric vector')
    if (length(pj) != ncri) stop('number of elements in prefferenceThreshold must be equal to number of criteria (set value to 0 for criteria with preference function that does not require this parameter).')
    for(i in 1:ncri){
      if (prefTypesVector[i] && pj[i] < 0){
        msg <- cat('prefference threshold for criterion ', i, ' set < 0, only non negative values accepted.')
        stop(msg)
      }
    }
  }
  if('Gaussian' %in% preferenceFunction){ # Gaussian type of prefference function check intermediate threshold for consistency
    if (is.null(sj) && is.null(pj) && is.null(qj)) stop('intermediate threshold required (and not possible to derive from indiference and prefference thresholds).')
    if (is.null(sj)) { # try to derive sj from other thresholds
      if (!is.numeric(pj) || !is.numeric(qj) || length(pj) != ncri || length(qj) != ncri) stop('intermediate threshold required (failed to derive it form other thresholds due to its inconsistencies)')
      sj <- (pj + qj)/2
      if (length(sj[sj < 0]) > 0) stop('intermediate threshold required (failed to derive it form other thresholds due to its inconsistencies)')
    }
  }
  if(!is.numeric(w) || length(w) != ncri) stop('weights not set for all criteria.')
  #if(sum(w) != 1) stop('sum of weights does not equal to 1.')

  #clean-up (remove variables used only for consistency check)
  fTypes     <- NULL
  indifTypes <- NULL
  prefTypes  <- NULL
  indifTypesVector <- NULL
  prefTypesVector  <- NULL
  ## End of checking the validity of the "inputs"

  nalt <- nrow(PM)  #no. of alternatives
  alt  <- rownames(PM) #list of alternatives
  cri  <- colnames(PM) #list of criteria

  #pairwaise comparison (validated)
  DK <- list()
  for(k in 1:ncri) {
    DKf <- matrix(data = 0, nrow = nalt, ncol = nalt)
    for(i in 1:nalt) {
      for(j in 1:nalt){ #DK(ai,aj) = fK(ai) - fK(aj)
        DKf[i,j] <- PM[i,k] - PM[j,k]
      }
    }
    rownames(DKf) <- alt
    colnames(DKf) <- alt
    DK[[k]] <- DKf
  }
  names(DK) <- cri
  #print(DK) #DEBUG info

  #preference degree (validated)
  Pj <- list()
  for(k in 1:ncri){
    PjK <- matrix(data = 0, nrow = nalt, ncol = nalt)
    DKf <- DK[[k]]
    for(i in 1:nalt){
      for(j in 1:nalt){
        if(preferenceFunction[k] == 'default'){
          if(DKf[i,j] <= 0){
            PjK[i,j] <- 0
          }else{
            PjK[i,j] <- 1
          }
        }else if(preferenceFunction[k] == 'U-shape'){
          if(DKf[i,j] <= qj[k]){
            PjK[i,j] <- 0
          }else{
            PjK[i,j] <- 1
          }
        }else if(preferenceFunction[k] == 'V-shape'){
          if(DKf[i,j] <= 0){
            PjK[i,j] <- 0
          }else if(DKf[i,j] > 0 & DKf[i,j] <= pj[k]){
            PjK[i,j] <- DKf[i,j]/pj[k]
          }else{
            PjK[i,j] <- 1
          }
        }else if(preferenceFunction[k] == 'level'){
          if(DKf[i,j] <= qj[k]){
            PjK[i,j] <- 0
          }else if(DKf[i,j] > qj[k] & DKf[i,j] <= pj[k]){
            PjK[i,j] <- 0.5
          }else{
            PjK[i,j] <- 1
          }
        }else if(preferenceFunction[k] == 'linear'){
          if(DKf[i,j] <= qj[k]){
            PjK[i,j] <- 0
          }else if(DKf[i,j] > qj[k] & DKf[i,j] <= pj[k]){
            PjK[i,j] <- (DKf[i,j] - qj[k])/(pj[k] - qj[k])
          }else{
            PjK[i,j] <- 1
          }
        }else if(preferenceFunction[k] == 'Gaussian'){
          if(DKf[i,j] <= 0){
            PjK[i,j] <- 0
          }else{
            PjK[i,j] <- 1 - exp(-((DKf[i,j]*DKf[i,j])/(2*sj[k]*sj[k])))
          }
        }
      }
    }
    rownames(PjK) <- alt
    colnames(PjK) <- alt
    Pj[[k]] <- PjK
  }
  names(Pj) <- cri
  #print(Pj) #DEBUG info

  #criteria flows (validated)
  pf_cri  <- matrix(data = 0, nrow = nalt, ncol = ncri)
  nf_cri  <- matrix(data = 0, nrow = nalt, ncol = ncri)
  net_cri <- matrix(data = 0, nrow = nalt, ncol = ncri)
  for(k in 1:ncri){
    for(i in 1:nalt){
      t <- Pj[[k]]
      pf_cri[i, k] <- sum(t[i,])/(nalt - 1)
      nf_cri[i, k] <- sum(t[,i])/(nalt - 1)
    }
  }
  t <- NULL
  net_cri <- pf_cri - nf_cri
  colnames(pf_cri)  <- cri
  colnames(nf_cri)  <- cri
  colnames(net_cri) <- cri
  rownames(pf_cri)  <- alt
  rownames(nf_cri)  <- alt
  rownames(net_cri) <- alt
  #print("positive flow - criteria") #DEBUG info
  #print(pf_cri)
  #print("negative flow - criteria")
  #print(nf_cri)
  #print("net flow - criteria")
  #print (net_cri)

  #weighted unicriteria flows (validated)
  pf_cri_w <- sweep(pf_cri, MARGIN=2, w, `*`)
  nf_cri_w <- sweep(nf_cri, MARGIN=2, w, `*`)
  net_cri_w <- pf_cri_w - nf_cri_w
  colnames(pf_cri_w)  <- cri
  colnames(nf_cri_w)  <- cri
  colnames(net_cri_w) <- cri
  rownames(pf_cri_w)  <- alt
  rownames(nf_cri_w)  <- alt
  rownames(net_cri_w) <- alt
  #print("weighted positive flow - criteria") #DEBUG info
  #print(pf_cri_w)
  #print("weighted negative flow - criteria")
  #print(nf_cri_w)
  #print("weighted net flow - criteria")
  #print (net_cri_w)

  #global preference flows
  pf <- rowSums(pf_cri_w)
  nf <- rowSums(nf_cri_w)
  netf <- pf - nf
  names(pf) <- alt
  names(nf) <- alt
  names(netf) <- alt
  #print("global positive flow") #DEBUG info
  #print(pf)
  #print("global negative flow")
  #print(nf)
  #print("global net flow")
  #print(netf)

  out <- list(
    positiveFlowCriteria = pf_cri,
    negativeFlowCriteria = nf_cri,
    netFlowCriteria = net_cri,
    weightedPositiveFlowCriteria = pf_cri_w,
    weightedNegativeFlowCriteria = nf_cri_w,
    weightedNetFlowCriteria = net_cri_w,
    positiveFlow = pf,
    negativeFlow = nf,
    netFlow = netf,
    preferenceDegreeUnw = Pj,
    pairweiseComparison = DK
  )
  return(out)
}
