# implementation of PROMETHEE III function. Function provides preordering by intervals.
#
# parameters
#   PM - performance matrix
#   preferenceFunction - vector of preference functions types to derive preference when comparing PM
#   w - weights
#   minmax - min/max value or vector of mixed values for criteria orientation
#   indifferenceTreshold list of indifference thresholds
#   prefferenceThreshold list of prefference thresholds
#   intermediateThreshold lish of intermediate thresholds for Gaussian function type
PROMETHEE_III <- function(PM, preferenceFunction, w, minmax = 'max',
                        indifferenceTreshold = NULL, prefferenceThreshold = NULL,
                        intermediateThreshold = NULL){
  #params consistency check centralized in generalized PROMETHEE function
  nalt <- nrow(PM)  #no. of alternatives
  alt  <- rownames(PM) #list of alternatives
  #here is only minmax evaluation
  PM <- util_pm_minmax(PM, minmax) #validate minmax and invert scales if neccessary
  #end of parameter consistency check

  flow <- PROMETHEE(PM, preferenceFunction, w, indifferenceTreshold, prefferenceThreshold, intermediateThreshold)
  pf <- flow$positiveFlow
  nf <- flow$negativeFlow
  nof <- pf - nf #net outranking flow (nof)
  names(nof) <- alt

  stand_error <- round((sd(nof)/sqrt(nalt)), 3)
  x_limit <- c()
  y_limit <- c()
  for (i in 1:nalt) {
    x <- round((nof[i] - stand_error), 3)
    y <- round((nof[i] + stand_error), 3)
    x_limit <- append(x_limit, x)
    y_limit <- append(y_limit, y)
  }
  #establish preference system
  pref <- matrix(data = '-', nrow = nalt, ncol = nalt)
  for (i in 1:nalt) {
    for (j in 1:nalt) {
      if (i != j){
        if (x_limit[i] > y_limit[j]){
          pref[i,j] <- 'P+'
        } else if ((x_limit[i] <= y_limit[j]) && (x_limit[j] <= y_limit[i])){
          pref[i,j] <- 'I'
        } else {
          pref[i,j] <- 'p-'
        }
      }
    }
  }
  rownames(pref) <- alt
  colnames(pref) <- alt

  out <- list(
    positiveFlow = pf,
    negativeFlow = nf,
    netFlow = nof,
    preferenceDegree = flow$preferenceDegree,
    preferenceDegreeUnw = flow$preferenceDegreeUnw,
    pairweiseComparison = flow$pairweiseComparison,
    preferenceMatrix = pref
  )
  return(out)
}
